/**
 * 1998/f1040/f1040.c
 * 
 * This file is part of the taxpatch program
 * Copyright (C) 1999 Frederick (Rick) Niles
 * This code is covered under the GPL (see COPYING for details)
 **/


#define _POSIX_SOURCE 1

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include "taxpatch.h"

#define IRSFILE "f1040.ps"

static int filingStatus = 1;
static int exemptSelf=0, exemptSpouse=0, numDepList=0;
static double totalIncome = 0;
static double adjustedGrossIncome = 0;
static double earnedIncome = 0;

extern int dontPrintZeroValue;

void
pageOne1040(FILE *fp)
{
  char line[MAXLINELEN];
  int i, status;
  double dtmp;
  struct stat buf;
  time_t tim;

  char firstName[MAXLINELEN];
  char initial;
  char lastName[MAXLINELEN];
  char SSN[MAXLINELEN];
  char spouseFirstName[MAXLINELEN];
  char spouseInitial;
  char spouseLastName[MAXLINELEN];
  char spouseSSN[MAXLINELEN];
  char homeAddress[MAXLINELEN];
  char aptNo[MAXLINELEN];
  char cityTown[MAXLINELEN];
  int contrib=0, spouseContrib=0;

  char spouseFullName[MAXLINELEN];
  char childName[MAXLINELEN];
  int deathYear, jointReturn=0;

  int numChildWith=0, numChildSep=0, depQualCredit[6], totalQualCredit=0;
  char deps[5][6][MAXLINELEN];

  char alimonySSN[MAXLINELEN];
  int hunkcount;
  double valuebuf[30], line32;

  /* Read Label section */

  readNextLine(fp, firstName);
  readNextLine(fp, line);
  initial = line[0];
  readNextLine(fp, lastName);
  readNextLine(fp, SSN);

  readNextLine(fp, spouseFirstName);
  readNextLine(fp, line);
  spouseInitial = line[0];
  readNextLine(fp, spouseLastName);
  readNextLine(fp, spouseSSN);

  readNextLine(fp, homeAddress);
  readNextLine(fp, aptNo);
  if(strncmp("NA", aptNo, 2) == 0)
    aptNo[0] = 0;
  readNextLine(fp, cityTown);

  contrib = readBoolean(fp);
  spouseContrib = readBoolean(fp);

  /* read Filing Status section */
  readNextLine(fp, line);
  sscanf(line, "%d", &filingStatus);
  if(filingStatus == 2)
    jointReturn = 1;
  readNextLine(fp, spouseFullName);
  readNextLine(fp, childName);
  readNextLine(fp, line);
  sscanf(line, "%d", &deathYear);

  /* read Dependents section */
  exemptSelf   = readBoolean(fp);
  exemptSpouse = readBoolean(fp) 
    && ((filingStatus == 2) || (filingStatus == 3));

  readNextLine(fp, line);
  sscanf(line, "%d", &numDepList);
  for(i = 0; i < numDepList; i++)
    {
      readNextLine(fp, line);
      sscanf(line,"%s %s %s %s %s", 
	     deps[0][i], deps[1][i], deps[2][i],
	     deps[3][i], deps[4][i]);
      if((deps[4][i][0] == 'Y') || (deps[4][i][0] == 'y'))
	{
	  depQualCredit[i]=1;
	  totalQualCredit++;
	}
      else
	depQualCredit[i]=0;      
    }
  readNextLine(fp, line);
  sscanf(line, "%d", &numChildWith);
  readNextLine(fp, line);
  sscanf(line, "%d", &numChildSep);

  for(i = 0; i < 28; i++)
    {
      readNextLine(fp, line);
      sscanf(line, "%lf", &(valuebuf[i]));
    }
  readNextLine(fp, alimonySSN);

  printf("Value on line 31 is %.2f\n", valuebuf[27]);
  printf("SSN: %s\n", alimonySSN);
    
  /***********************************************************/
  /******** End of reading section ***************************/
  /***********************************************************/

  /**
   * This next section is neccessary "out of order", since the patch
   * file must be in line number ordering, and the form is not produced
   * top to bottom monotonically.
   *
   * It other words, it's a mess, live with it.
   **/

  /* Done reading, start the patch. */
  /* Get the last modification time of original IRS 1040 PS file. */
  status = stat(IRSFILE, &buf);
  if(status != 0)
    {
      perror("Could not stat() IRS file");
    }
  tim = time(NULL);

  /* Have the patch date mean something at least. */
  printf("*** f1040.ps    %s",
	 ctime(&(buf.st_ctime)));
  printf("--- f1040-%s_%s.ps      %s",
	 firstName, lastName,
	 ctime(&(tim)));

  /* SSNs first. */
  hunkHeader(2039, 3*(jointReturn+1));
  printf("  11.95 0 0 11.95 114.045 738.084 Tm\n"
	 "  0.04 Tw\n"
	 "  [ (U.S. Individual Income T) 70 (ax Retur) -10 (n) TJ\n"
	 "+  29.8 -3.1 TD\n"
	 "+ (%c%c%c  %c%c  %c%c%c%c) Tj\n",
	 SSN[0], SSN[1], SSN[2],
	 SSN[4], SSN[5],
	 SSN[7], SSN[8], SSN[9], SSN[10]);

  if(jointReturn)
    {
      printf("+ 0.0 -2.0 TD\n"
	     "+ (%c%c%c  %c%c  %c%c%c%c) Tj\n"
	     "+ 0.0  2.0 TD\n",
	     spouseSSN[0], spouseSSN[1], spouseSSN[2],
	     spouseSSN[4], spouseSSN[5],
	     spouseSSN[7], spouseSSN[8], spouseSSN[9], spouseSSN[10]);
    }

  printf("+ -29.8  3.1 TD\n"
	 "  ET\n"
	 "  0 J\n"
	 "  [] 0 d\n");
  

  /* Next the Filing Status */
  switch(filingStatus)
    {
    case 1:
      dtmp = 0.95;
      break;
    case 2:
      dtmp = 0.0;
      break;
    case 3:
      dtmp = -1.1;
      break;
    case 4:
      dtmp = -2.4;
      break;
    case 5:
      dtmp = -4.1;
      break;
    default:
      fprintf(stderr, "Unknown filingStatus value.\n");
      exit(EXIT_FAILURE);      
    }
  hunkHeader(4200, 3);
  printf("  11.95 0 0 11.95 41.826 581.636 Tm\n"
	 "  0.1 Tw\n"
	 "  (Filing Status) Tj\n"
	 "+  8.5  %f TD\n"
	 "+ (X) Tj\n"
	 "+ -8.5  %f TD\n"
	 "  _56 1 Tf\n"
	 "  7.967 0 0 7.967 118.407 579.246 Tm\n"
	 "  (2) Tj\n", dtmp, -dtmp);

  if(filingStatus == 5)
    {
      char ctmp[MAXLINELEN];
      sprintf(ctmp, "%4d", deathYear);
      substitute(4835,  
		 "  _50 1 Tf\n"
		 "  7.967 0 0 7.967 396.454 533.935 Tm\n"
		 "  0.006 Tc\n"
		 "! (", 
		 "19", ctmp, 
		 ") Tj\n"
		 "  4.118 0 TD\n"
		 "  0 Tc\n"
		 "  0.069 Tw\n");
	     
    }

  if(exemptSelf)
    {
      hunkHeader(4848, 3);
      printf("  (5) Tj\n"
	     "  0 -1.663 TD\n"
	     "  [ (6a) -2920 (Yourself. ) TJ\n"
	     "+  2.4 0.0 TD\n"
	     "+ (X) Tj\n"
	     "+ -2.4 0.0 TD\n"
	     "  _50 1 Tf\n"
	     "  8.6 0 TD\n"
	     "  0.033 Tw\n");
    }
  

      hunkHeader(4874, 15);
      printf("  _58 1 Tf\n"
	     "  11.95 0 0 11.95 42.622 510.134 Tm\n"
	     "  (Exemptions) Tj\n"
	     "+  42.7 0.0 TD\n"
	     "+ (%d) Tj \n"
	     "+ 0.0 -3.3 TD\n"
	     "+ (%d) Tj \n"
	     "+ 0.0 -2.8 TD\n"
	     "+ (%d) Tj \n"
	     "+ 0.0 -1.5 TD\n"
	     "+ (%d) Tj \n"
	     "+ 0.0 -1.9 TD\n"
	     "+ (%d) Tj \n"
	     "+ 0.0  1.9 TD\n"	     
	     "+ 0.0  1.5 TD\n"	     
	     "+ 0.0  2.8 TD\n"	     
	     "+ 0.0  3.3 TD\n"	     
	     "+ -42.7 0.0 TD\n"	     
	     "  _56 1 Tf\n"
	     "  7.967 0 0 7.967 150.672 497.935 Tm\n"
	     "  -0.001 Tc\n",
	     exemptSelf+exemptSpouse, numChildWith, numChildSep, 
	     numDepList - numChildWith - numChildSep, 
	     exemptSelf + exemptSpouse + numDepList);



  if(exemptSpouse)
    {
      hunkHeader(4878, 3);
      printf("  7.967 0 0 7.967 150.672 497.935 Tm\n"
	     "  -0.001 Tc\n"
	     "  (Spouse) Tj\n"
	     "+ -1.7 0.0 TD\n"
	     "+ (X) Tj\n"
	     "+  1.7 0.0 TD\n"
	     "  ET\n"
	     "  184.78 497.935 m\n"
	     "  472.581 497.935 l\n");
    }


  if(numDepList)
    {
      hunkHeader(4973, 12*numDepList+3*totalQualCredit);
      printf("  0 Tc\n"
	     "  0.069 Tw\n"
	     "  (see page 19.) Tj\n");
      for(i = 0; i < numDepList; i++)
	{
	  dtmp = 3.2 - 1.5*i;
	  printf("+  13.0 %f TD\n"
		 "+  (%s) Tj\n"
		 "+   8.0 0.0 TD\n"
		 "+  (%s) Tj\n"
		 "+   10.5 0.0 TD\n"
		 "+  (%c%c%c   %c%c   %c%c%c%c) Tj\n"
		 "+   10.0 0.0 TD\n"
		 "+  (%s) Tj\n",
		 dtmp, 
		 deps[0][i], deps[1][i], 
		 deps[2][i][0], deps[2][i][1], deps[2][i][2],
		 deps[2][i][4], deps[2][i][5], 
		 deps[2][i][7], deps[2][i][8], deps[2][i][9], deps[2][i][10],
		 deps[3][i]);

	  if(depQualCredit[i])
	    {
	      printf("+   9.3 0.2 TD\n"
		     "+   (X) Tj\n"
		     "+   -9.3 -0.2 TD\n");
	    }

	  printf("+  -10.0 0.0 TD\n"
		 "+  -10.5 0.0 TD\n"
		 "+  -8.0 0.0 TD\n"
		 "+  -13.0 %f TD\n", -dtmp);

	}
      printf("  ET\n"
	     "  .setdefaultscreen\n"
	     "  false setstrokeadjust\n");
    }


  /* Values section */
  
  /**
   * Calculation section
   **/

  /* Line 22 Calculation: Total Income. */
  totalIncome = valuebuf[ 0] 
	  + valuebuf[ 1] 
	  + valuebuf[ 3] 
	  + valuebuf[ 4] 
	  + valuebuf[ 5] 
	  + valuebuf[ 6] 
	  + valuebuf[ 7] 
	  + valuebuf[ 8] 
	  + valuebuf[10] 
	  + valuebuf[12] 
	  + valuebuf[13] 
	  + valuebuf[14] 
	  + valuebuf[15] 
	  + valuebuf[17] 
	  + valuebuf[18];

  /* Line 32 Calculation: Lines 23 through 31a. */
  line32 = valuebuf[19] 
    + valuebuf[20] 
    + valuebuf[21] 
    + valuebuf[22] 
    + valuebuf[23] 
    + valuebuf[24] 
    + valuebuf[25] 
    + valuebuf[26] 
    + valuebuf[27];

  adjustedGrossIncome = totalIncome - line32;

  /* End calculation section */

  /**
   * Values patch section
   **/
  rowWidth = 1.507;
  colWidth = 13.6;
  hunkcount = 8;

  /* First check for zero values. */
  if(dontPrintZeroValue)
    {
      for(i = 0; i < 28; i++)
	if(valuebuf[i] != 0.0)
	  hunkcount += 3;
      hunkcount += 3*(line32!=0);
    }
  else
    hunkcount += 3*29;

  hunkcount += 3*(valuebuf[27]!=0);
  hunkHeader(5010, hunkcount);
  printf("  _56 1 Tf\n"
	 "  7.967 0 0 7.967 468.199 377.238 Tm\n"
	 "  (7) Tj\n");

  addRowColValue( 0, 0, valuebuf[0]); 
  addRowColValue( 1, 0, valuebuf[1]); 
  addRowColValue( 2,-1, valuebuf[2]); 
  for(i = 3; i < 9; i++)
    addRowColValue( i, 0, valuebuf[i]); 

  /* We need just a little tweak. */
  printf("+  -1.8 0.0 TD\n");
  addRowColValue( 9,-2, valuebuf[ 9]); 
  addRowColValue(10,-2, valuebuf[11]); 
  addRowColValue(14,-2, valuebuf[16]); 
  printf("+   1.8 0.0 TD\n");

  addRowColValue( 9, 0, valuebuf[10]); 
  addRowColValue(10, 0, valuebuf[12]); 
  addRowColValue(11, 0, valuebuf[13]); 
  addRowColValue(12, 0, valuebuf[14]); 
  addRowColValue(13, 0, valuebuf[15]); 
  addRowColValue(14, 0, valuebuf[17]); 
  addRowColValue(15, 0, valuebuf[18]); 

  for(i = 17; i < 26; i++)
    addRowColValue(i,-1, valuebuf[i+2]); 
  addRowColValue(16, 0, totalIncome); 
  addRowColValue(26, 0, line32); 
  addRowColValue(27, 0, adjustedGrossIncome); 

  if(valuebuf[27] != 0)
    {
      printf("+ %f %f TD\n"
	     "+ (%c%c%c    %c%c    %c%c%c%c) Tj\n"
	     "+ %f %f TD\n",
	     -25.0, -25*rowWidth,
	     alimonySSN[0], alimonySSN[1], alimonySSN[2],
	     alimonySSN[4], alimonySSN[5],
	     alimonySSN[7], alimonySSN[8], alimonySSN[9], alimonySSN[10],
	     25.0, -(-25*rowWidth));

    }


  printf("  ET\n"
	 "  0.498 w\n"
	 "  459.635 375.246 m\n");

      
  /* Spouse Full Name for Married filing separately section */
  if(filingStatus == 3)
    {
      hunkHeader(5785, 3);
      printf("  0 Tc\n"
	     "  0.067 Tw\n"
	     "  (Married filing separate return."
	     " Enter spouse's social security no. above"
	     " and full name here.) Tj\n"
	     "+ 44.0 0.3 TD\n"
	     "+ (%s) Tj\n"
	     "+ -44.0 -0.3 TD\n"
	     "  _64 1 Tf\n"
	     "  5.975 0 0 5.975 455.143 568.74 Tm\n"
	     "  (\\251) Tj\n",
	     spouseFullName);
    }

  /* Finally, the actual label area (Name, address...) */
  
  hunkHeader(6445, 16+10*jointReturn);
  printf("  13.942 0 0 13.942 41.826 713.387 Tm\n"
	 "  -0.002 Tc\n"
	 "  (Label) Tj\n"
	 "+ 6.0 -0.85 TD\n"
	 "+ (%s) Tj\n"
	 "+ 7.0 0.0 TD\n"
	 "+ (%c) Tj\n"
	 "+ 5.0 0.0 TD\n"
	 "+ (%s) Tj\n",
	 firstName, initial, lastName);

  if(jointReturn)
    printf("+ -12.0 -1.8 TD\n"
	   "+ (%s) Tj\n"
	   "+ 7.0 0.0 TD\n"
	   "+ (%c) Tj\n"
	   "+ 5.0 0.0 TD\n"
	   "+ (%s) Tj\n"
	   "+ 0.0 1.8 TD\n",
	   spouseFirstName, spouseInitial, spouseLastName);


  printf("+ -12.0 -3.5 TD\n"
	 "+ (%s) Tj\n"
	 "+  20.0 0.0 TD\n"
	 "+ (%s) Tj\n"
	 "+ -20.0 0.0 TD\n"
	 "+ 0.0 -1.6 TD\n"
	 "+ (%s) Tj\n"
	 "+ %f -1.0 TD\n"
	 "+ (X) Tj\n"
	 "+ -%f 1.0 TD\n",
	 homeAddress, aptNo, cityTown, 
	 26.0-1.5*contrib,
	 26.0-1.5*contrib);

  if(jointReturn)
    printf("+ %f -1.9 TD\n"
	   "+ (X) Tj\n"
	   "+ -%f 1.9 TD\n",
	   26.0-1.5*spouseContrib,
	   26.0-1.5*spouseContrib);

  printf("  ET\n"
	 "  337.235 255.256 m\n"
	 "  337.235 277.404 l\n");


  if(filingStatus == 4)
    {
      hunkHeader(6488, 3);
      printf("  0 -1.375 TD\n"
	     "  0.069 Tw\n"
	     "  (enter this child's name here. ) Tj\n"
	     "+  15.0  0.1 TD\n"
	     "+ (%s) Tj\n"
	     "+ -15.0 -0.1 TD\n"
	     "  _64 1 Tf\n"
	     "  5.975 0 0 5.975 269.925 545.736 Tm\n"
	     "  (\\251) Tj\n", childName);
	     
    }


}

double
taxRateSchedule(double taxableIncome, double *breakpt)
{
  double tax = 0;
  double tmp = taxableIncome;
  double percent[5] = {0.15, 0.13, 0.03, 0.05, 0.036};
  int i;

  for(i = 0; i < 5; i++)
    {
      tax += percent[i] * tmp;
      tmp -= breakpt[i];
      tmp = max(tmp, 0.0);
    }

  /* Round to the nearest dollar. */
  tax = floor(tax+0.5);

  return tax;
}
double
calculateTax(double taxableIncome)
{
  double tax=-1;

  /* Round the funny ranges used in the tax chart. */
  if(taxableIncome < 5)
    tax = 0;
  else if(taxableIncome < 15)
    tax = 2;
  else if(taxableIncome < 25)
    tax = 3;
  else if(taxableIncome < 3000)
    taxableIncome = (floor((4*taxableIncome)/100)*100+50)/4.0;
  else if(taxableIncome < 100000)
    taxableIncome = (floor((2*taxableIncome)/100)*100+50)/2.0;

  /* Check if we need to calcuate tax and setup break points. */
  if(tax < 0)
    {
      double breakpt[5];
      switch(filingStatus)
	{
	case 1:
	  breakpt[0] = 25350;
	  breakpt[1] = 61400;
	  breakpt[2] = 128100;
	  breakpt[3] = 278450;
	  break;
	case 2:
	case 5:
	  breakpt[0] = 42350;
	  breakpt[1] = 102300;
	  breakpt[2] = 155950;
	  breakpt[3] = 278450;
	  break;
	case 3:
	  breakpt[0] = 21175;
	  breakpt[1] = 51150;
	  breakpt[2] = 77975;
	  breakpt[3] = 139225;
	  break;
	case 4:
	  breakpt[0] = 33950;
	  breakpt[1] = 87700;
	  breakpt[2] = 142000;
	  breakpt[3] = 278450;
	  break;

	}
      tax = taxRateSchedule(taxableIncome, breakpt);
    }
  return tax;
}

void
pageTwo1040(FILE *fp)
{
  char line[MAXLINELEN];
  int i;
  int tmp;
  double dtmp;

  int age65=0, blind=0, spouseAge65=0, spouseblind=0, box35a=0, box35b=0;
  int taxForm8814=0, taxForm4972=0;
  int taxForm3800=0, taxForm8396=0, taxForm8801=0, box47d=0;
  int taxForm2439=0, taxForm4136=0;
  int directdeposit=0, ddAccountType=0;
  int paidPreparer=0, selfEmployed=0;
  double itemizedDeductions;
  char taxFormOther[MAXLINELEN];
  char EICtype[MAXLINELEN];
  char ddRoutingNumber[MAXLINELEN];
  char ddAccountNumber[MAXLINELEN];
  char occupation[MAXLINELEN];
  char occupationSpouse[MAXLINELEN];
  char phoneNumber[MAXLINELEN];
  char preparerSSN[MAXLINELEN];
  char preparerEIN[MAXLINELEN];
  char preparerZIP[MAXLINELEN];
  char preparerAddress[2][MAXLINELEN];
  char strbuf[MAXLINELEN];
  int refundDue=0;

  int hunkcount;
  double valuebuf[30], desiredRefund;
  double deduction, tax, taxableIncome;
  double line37, line38, totalCredits, line49;
  double totalTax, totalPayments;

  /* Read tax file first. */
  /* Tax and Credits section */

  age65       = readBoolean(fp);
  blind       = readBoolean(fp);
  spouseAge65 = readBoolean(fp)
    && ((filingStatus == 2) || (filingStatus == 3));
  spouseblind = readBoolean(fp)
    && ((filingStatus == 2) || (filingStatus == 3));
  box35b      = readBoolean(fp);
  readNextLine(fp, line);
  sscanf(line, "%lf", &itemizedDeductions);
  taxForm8814 = readBoolean(fp);
  taxForm4972 = readBoolean(fp);
  readNextLine(fp, line);
  sscanf(line, "%lf", &tax);
  for(i = 0; i < 7; i++)
    {
      readNextLine(fp, line);
      sscanf(line, "%lf", &(valuebuf[i]));
    }
  readNextLine(fp, line);
  sscanf(line, "%d", &(tmp));
  for(i = 0; i < tmp; i++)
    {
      readNextLine(fp, line);
      if(!strncmp("3800", line, strlen("3800")))
	taxForm3800 = 1;
      else if(!strncmp("8396", line, strlen("8396")))
	taxForm8396 = 1;
      else if(!strncmp("8801", line, strlen("8801")))
	taxForm8801 = 1;
      else
	{
	  box47d = 1;
	  strncpy(taxFormOther, line, MAXLINELEN);
	}
    }

  for(i = 7; i < 17; i++)
    {
      readNextLine(fp, line);
      sscanf(line, "%lf", &(valuebuf[i]));
    }

  readNextLine(fp, EICtype);
  for(i = 17; i < 21; i++)
    {
      readNextLine(fp, line);
      sscanf(line, "%lf", &(valuebuf[i]));
    }
  readNextLine(fp, line);
  if(valuebuf[20] != 0)
    {
      sscanf(line, "%c %c", &(strbuf[0]), &(strbuf[1]));
      if((strbuf[0] == 'a') || (strbuf[1] == 'a'))
	taxForm2439 = 1;
      if((strbuf[0] == 'b') || (strbuf[1] == 'b'))
	taxForm4136 = 1;
    }

  /* begin Refund read section */

  readNextLine(fp, line);
  sscanf(line, "%lf", &(desiredRefund));
  directdeposit = readBoolean(fp);
  readNextLine(fp, ddRoutingNumber);
  readNextLine(fp, line);
  ddAccountType = strncmp("Savings", line, strlen("Savings"));
  readNextLine(fp, ddAccountNumber);
  readNextLine(fp, line);
  sscanf(line, "%lf", &(valuebuf[21])); 
  readNextLine(fp, line);
  sscanf(line, "%lf", &(valuebuf[22]));
  readNextLine(fp, occupation);
  readNextLine(fp, occupationSpouse);
  readNextLine(fp, phoneNumber);

  /* Paid Preparer's section */
  paidPreparer = readBoolean(fp);
  selfEmployed = readBoolean(fp);
  readNextLine(fp, preparerSSN);
  readNextLine(fp, preparerEIN);
  readNextLine(fp, preparerAddress[0]);
  readNextLine(fp, preparerAddress[1]);
  readNextLine(fp, preparerZIP);

  /* End of reading section. */
	       

  /**
   * Calculation section
   **/

  /* Calculate deduction */
  box35a = age65 + blind + (spouseAge65 + spouseblind)*(exemptSpouse);
  switch(filingStatus)
    {
    case 1:
      deduction = 4250;
      dtmp = 1050;
      break;
    case 2:
    case 5:
      deduction = 7100;
      dtmp = 850;
      break;
    case 3:
      deduction = 3500;
      dtmp = 850;
      break;
    case 4:
      deduction = 6250;
      dtmp = 1050;
      break;
    }
  /* Special case: Standard Deduction Worksheet for Dependents, page 29 */
  if(!exemptSelf)
    {
      deduction = min(deduction, max(700.0, earnedIncome + 250.0));
    }
  deduction += box35a*dtmp;
  deduction = max(deduction, itemizedDeductions);
  line37 = adjustedGrossIncome - deduction;
  if(adjustedGrossIncome < 93400)
    line38 = 2700*(exemptSelf+ exemptSpouse + numDepList);
  else
    {
      /* Deduction for Exemptions Worksheet -- Line 38, page 30 */
      double step[10];
      step[2] = 2700*(exemptSelf+ exemptSpouse + numDepList);
      step[3] = adjustedGrossIncome;
      switch(filingStatus)
	{
	case 1:
	  step[4] = 124500;
	  break;
	case 2:
	case 5:
	  step[4] = 186800;
	  break;
	case 3:
	  step[4] = 93400;
	  break;
	case 4:
	  step[4] = 155650;
	  break;
	}
      step[5] = step[3] - step[4];
      if(step[5] < 0)
	line38 = step[2];
      else if(((step[5] > 122500) && 
	      (filingStatus =! 2)) ||
	      ((step[5] > 61250) &&
	       (filingStatus == 2)))
	line38 = 0;
      else
	{
	  if(filingStatus == 2)
	    step[6] = ceil(step[5] / 1250);
	  else
	    step[6] = ceil(step[5] / 2500);
	  step[7] = 0.02 * step[6];
	  step[8] = step[2] * step[7];
	  line38 = step[2] - step[8];
	}
    }

  taxableIncome = max(line37 - line38, 0);

  /* Calculate Tax */
  if(tax == 0)
    tax = calculateTax(taxableIncome);
  
  totalCredits = valuebuf[0] 
    + valuebuf[1]
    + valuebuf[2]
    + valuebuf[3]
    + valuebuf[4]
    + valuebuf[5]
    + valuebuf[6];

  line49 = max(tax - totalCredits, 0);

  totalTax = valuebuf[7]
    + valuebuf[8]
    + valuebuf[9]
    + valuebuf[10]
    + valuebuf[11]
    + valuebuf[12]
    + line49;

  totalPayments = valuebuf[13]
    + valuebuf[14]
    + valuebuf[15]
    + valuebuf[17]
    + valuebuf[18]
    + valuebuf[19]
    + valuebuf[20];

  refundDue = totalPayments > totalTax;

  /* End of calculation section */


  hunkHeader(8670, 3);
  printf("  _53 1 Tf\n"
	 "  9.958 0 0 9.958 570.453 751.229 Tm\n"
	 "  (2) Tj\n"
	 "+ -11.7 -3.1 TD\n"
	 "+  (%d) Tj\n"
	 "+  11.7  3.1 TD\n"
	 "  _51 1 Tf\n"
	 "  6.672 0 0 6.971 50.002 751.229 Tm\n"
	 "  0 Tc\n", box35a);


  /**
   * Values patch section
   **/
  hunkcount = 3+4+3*refundDue;

  /* First check for zero values. */
  if(dontPrintZeroValue)
    {
      for(i = 0; i < 21; i++)
	if(valuebuf[i] != 0.0)
	  hunkcount += 3;
      if(valuebuf[22-refundDue] != 0.0)
	hunkcount += 3;      
      if(adjustedGrossIncome != 0.0)
	hunkcount += 3;
      if(deduction != 0.0)
	hunkcount += 3;
      if(line37 != 0.0)
	hunkcount += 3;
      if(line38 != 0.0)
	hunkcount += 3;
      if(taxableIncome != 0.0)
	hunkcount += 3;
      if(tax != 0.0)
	hunkcount += 3;
      if(totalCredits != 0.0)
	hunkcount += 3;
      if(line49 != 0.0)
	hunkcount += 3;
      if(totalTax != 0.0)
	hunkcount += 3;
      if(totalPayments != 0.0)
	hunkcount += 3;
    }
  else
    hunkcount += 3*32;

  hunkHeader(8699, hunkcount);
  rowWidth = 1.507;
  colWidth = 13.4;
  printf("  (34) Tj\n"
	 "  45.93 0.25 TD\n"
	 "  (34) Tj\n");


  /* We need just a little tweak. */
  printf("+  0.2 0.0 TD\n");

  addRowColValue(12,-1, valuebuf[ 0]); 
  addRowColValue(13,-1, valuebuf[ 1]); 
  addRowColValue(14,-1, valuebuf[ 2]); 
  addRowColValue(15,-1, valuebuf[ 3]); 
  addRowColValue(16,-1, valuebuf[ 4]); 
  addRowColValue(17,-1, valuebuf[ 5]); 
  addRowColValue(19,-1, valuebuf[ 6]); 
  addRowColValue(22, 0, valuebuf[ 7]); 
  addRowColValue(23, 0, valuebuf[ 8]); 
  addRowColValue(24, 0, valuebuf[ 9]); 
  addRowColValue(25, 0, valuebuf[10]); 
  addRowColValue(26, 0, valuebuf[11]); 
  addRowColValue(27, 0, valuebuf[12]); 
  addRowColValue(29,-1, valuebuf[13]); 
  addRowColValue(30,-1, valuebuf[14]); 
  addRowColValue(33,-1, valuebuf[15]); 
  printf("+ -1.1 0.0 TD\n");
  addRowColValue(32,-2, valuebuf[16]); 
  printf("+  1.1 0.0 TD\n");
  addRowColValue(34,-1, valuebuf[17]); 
  addRowColValue(35,-1, valuebuf[18]); 
  addRowColValue(36,-1, valuebuf[19]); 
  addRowColValue(37,-1, valuebuf[20]); 
  if(refundDue)
    addRowColValue(44,-1, valuebuf[21]); 
  else
    addRowColValue(47,-1, valuebuf[22]); 
  addRowColValue( 0, 0, adjustedGrossIncome); 
  addRowColValue( 6, 0, deduction); 
  addRowColValue( 7, 0, line37);
  addRowColValue( 9, 0, line38);
  addRowColValue(10, 0, taxableIncome);
  addRowColValue(11, 0, tax);
  addRowColValue(20, 0, totalCredits);
  addRowColValue(21, 0, line49);
  addRowColValue(28, 0, totalTax);
  addRowColValue(38, 0, totalPayments);

  if(refundDue)
    {
      tmp = 39;
      dtmp = totalPayments - totalTax;
      desiredRefund = min(desiredRefund, dtmp);
      desiredRefund -= valuebuf[21];
    }
  else
    {
      tmp = 46;
      dtmp = totalTax - totalPayments;
      dtmp += valuebuf[22];
    }
  addRowColValue(tmp, 0, dtmp);
  if(refundDue)
    addRowColValue(40, 0, desiredRefund);
  printf("+ -0.2 0.0 TD\n");

  printf("  ET\n"
	 "  0 J\n"
	 "  0.498 w\n");

  /* End of values patch section */

  /**
   * Check Box section 
   **/
  tmp = age65 + blind + spouseAge65 + spouseblind + box35b
    + taxForm8814 + taxForm4972
    + taxForm3800 + taxForm8396 + taxForm8801 + 2*box47d
    + taxForm2439 + taxForm4136 + (refundDue && directdeposit)
    + (selfEmployed && paidPreparer);
  if(tmp)
    {
      hunkHeader(8723, 3*tmp);
      printf("  false setoverprint\n"
	     "  0.002 Tc\n"
	     "  (35a) Tj\n");

      if(age65)
	addCheckMark(7.4, 0.12);
      if(blind)
	addCheckMark(20.05, 0.12);
      if(spouseAge65)
	addCheckMark(24.6, 0.12);
      if(spouseblind)
	addCheckMark(38.1, 0.12);
      if(box35b)
	addCheckMark(42.66, -3.9);

      if(taxForm8814)
	addCheckMark(21.9, -14.95);
      if(taxForm4972)
	addCheckMark(32.7, -14.95);

      if(taxForm3800)
	addCheckMark(13.7, -25.5);
      if(taxForm8396)
	addCheckMark(23.65, -25.5);
      if(taxForm8801)
	addCheckMark(3.7, -27);
      if(box47d)
	{
	  addCheckMark(13.7, -27);
	  printf("+ 21.5 -27 TD\n"
		 "+ (%s) Tj\n"
		 "+ -21.5 27 TD\n", taxFormOther);
	}

      if(taxForm2439)
	addCheckMark(17.3, -54);
      if(taxForm4136)
	addCheckMark(24.6, -54);

      if(refundDue && directdeposit)
	{
	  if(ddAccountType)
	    addCheckMark(32.7, -61.4); /* Checking */
	  else
	    addCheckMark(39.05, -61.4); /* Savings */
	}

      if(selfEmployed && paidPreparer)
	addCheckMark(43.4, -80.9);

      printf("  11.95 0 0 11.95 50.002 729.44 Tm\n"
	     "  0 Tc\n"
	     "  0.052 Tw\n");
    }  
  /* End of Check Box section */

  /**
   * Begin Direct Deposit Account section 
   **/
  if(refundDue && directdeposit)
    {
      hunkHeader(9831, 3*(9+17));
      printf("  45.696 -4.395 TD\n"
	     "  0.002 Tc\n"
	     "  (66a) Tj\n");
      for(i = 0; i < 9; i++)
	{
	  dtmp = -34.44+1.82*i;
	  printf("+  %f -2.6 TD\n"
		 "+ (%c) Tj\n"
		 "+  %f  2.6 TD\n",
		 dtmp, ddRoutingNumber[i], -dtmp);
	}
      for(i = 0; i < 17; i++)
	{
	  dtmp = -34.44+1.82*i;
	  printf("+  %f -4.3 TD\n"
		 "+ (%c) Tj\n"
		 "+  %f  4.3 TD\n",
		 dtmp, ddAccountNumber[i], -dtmp);
	}
      printf("  -45.696 -0.249 TD\n"
	     "  (66a) Tj\n"
	     "  ET\n"); 
    }
  /* End Direct Deposit Account section */

  /** 
   * Sign Here section
   **/

  tmp = 4 + (filingStatus == 2) + (paidPreparer)*12;
  hunkHeader(11441, tmp);
  preparerEIN[2] = 0;
  printf("  6.971 0 0 6.971 374.211 142.247 Tm\n"
	 "  0.07 Tw\n"
	 "  (Your occupation) Tj\n"
	 "+ 0 -2.3 TD\n"
	 "+ (%s) Tj\n"
	 "+ 0 -3.5 TD\n", occupation);
  if(filingStatus == 2)
    printf("+ (%s) Tj\n",occupationSpouse);  
  if(paidPreparer)
    printf("+ 15 -3.5 TD\n"
	   "+ (%s) Tj\n"
	   "+  3 -1.8 TD\n"
	   "+ (%s) Tj\n"
	   "+  2 0 TD\n"
	   "+ (%s) Tj\n"
	   "+ -45 0 TD\n"
	   "+ (%s) Tj\n"
	   "+  0 -1.7 TD\n"
	   "+ (%s) Tj\n"
	   "+  45  0 TD\n"
	   "+ (%s) Tj\n"
	   "+  -20 12.8 TD\n",
	   preparerSSN,
	   preparerEIN, &(preparerEIN[3]), preparerAddress[0],
	   preparerAddress[1], preparerZIP);
  else
    printf("+  0 5.8 TD\n");
  printf("  0 -3.871 TD\n"
	 "  0.075 Tw\n"
	 "  (Spouse's occupation) Tj\n");

  if(strcmp(phoneNumber, "none"))
    {
      phoneNumber[4] = 0;
      printf("***************\n"
	     "*** 11863,11870 ****\n"
	     "  0.063 Tw\n"
	     "  (number \\(optional\\)) Tj\n"
	     "  0 -4.454 TD\n"
	     "! 3.101 Tc\n"
	     "! (\\(\\)) Tj\n"
	     "  ET\n"
	     "  367.001 147.216 m\n"
	     "  367.001 99.266 l\n"
	     "--- 11863,11870 ----\n"
	     "  0.063 Tw\n"
	     "  (number \\(optional\\)) Tj\n"
	     "  0 -4.454 TD\n"
	     "! 0.1 Tc\n"
	     "! (\\(%s\\) %s) Tj\n"
	     "  ET\n"
	     "  367.001 147.216 m\n"
	     "  367.001 99.266 l\n", &(phoneNumber[1]),
	     &(phoneNumber[6]));
    }


  /* End of Sign Here section */

}


void
patchForm(FILE *fp)
{
  pageOne1040(fp);
  pageTwo1040(fp);
}
